/*
 *  Copyright (C) 2003-2006 Daniel Robert Bradley. All rights reserved.
 *
 *  This software is redistributable under the terms of the GNU LGPL
 *  (Lesser General Public License) Version 2.1, February 1999.
 *
 *  I.S.Labs is a registered trademark of Daniel Robert Bradley
 */

#ifdef __cplusplus
#include <openocl/base/psa/CRuntime.h>
#include <openocl/base/String.h>
#include <openocl/base/Testing.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#define true  1
#define false 0

/*************************************************************************
 *   Declarations
 *************************************************************************/

/*
 *  Constants
 */
#define ACTIVATE      0
#define REMOVE        1
#define TEST_ACTIVATE 2 
#define TEST_REMOVE   4 

#ifdef __cplusplus
using namespace openocl::base;
using namespace openocl::base::psa;
#else
	static void* null = (void*) 0;
	#ifndef BOOLEAN
	#define BOOLEAN
	typedef int bool;
	#endif
#endif

/*
 *  Argument variables.
 */

static bool all        = false;
static bool deactivate = false;
static bool nothing = false;

static bool bin     = false;
static bool etc     = false;
static bool include = false;
static bool lib     = false;
static bool man     = false;
static bool sbin    = false;
static bool texinfo = false;

static bool verbose = false;

/*
 *  Function prototypes.
 */

static bool activate();

static bool activate_packages( int count, char** packages );

static bool activate_package( const char* absolutePackagePath );

static bool activate_directory( const char* absolutePackagePath,
                                const char* absoluteTargetPath,
                                const char* directoryToActivate );

static bool activate_lib_directory( const char* absolutePackagePath,
                                    const char* absoluteTargetPath,
                                    const char* directoryToActivate );

static bool activate_man_directory( const char* absolutePackagePath,
                                    const char* absoluteTargetPath,
                                    const char* directoryToActivate );

static bool activate_file( const char* absolutePackagePath,
                           const char* absoluteTargetPath,
                           const char* directoryToActivate,
                           const char* file );

static bool deactivate_package( const char* app );

static bool deactivate_directory( const char* absolutePackagePath,
                                  const char* absoluteTargetPath,
                                  const char* directoryToDeactivate );

static bool deactivate_lib_directory( const char* absolutePackagePath,
                                      const char* absoluteTargetPath,
                                      const char* directoryToDeactivate );

static bool deactivate_man_directory( const char* absolutePackagePath,
                                      const char* absoluteTargetPath,
                                      const char* directoryToActivate );

static bool deactivate_file( const char* absolutePackagePath,
                             const char* absoluteTargetPath,
                             const char* directoryToActivate,
                             const char* file );

static char* find_absolute_path( const char* aRelativePath );

static char* find_cwd();

static char* find_target_base( const char* anAbsolutePath );

static int process_arguments( int argc, char** argv );

static void print_arguments();

static char* concat_dirs( const char* first, const char* second );

char* find_relative_path( const char* source, const char* target );

static int    ARGC;
static char** ARGV;

/*************************************************************************
 *   Tests
 *************************************************************************/

#ifndef OPENOCL_BASE_PSA_CRUNTIME_MALLOC
#define OPENOCL_BASE_PSA_CRUNTIME_MALLOC
void* CRuntime_malloc( int size )
{
	return malloc( size );
}
#endif

#ifndef OPENOCL_BASE_PSA_CRUNTIME_CALLOC
#define OPENOCL_BASE_PSA_CRUNTIME_CALLOC
void* CRuntime_calloc( unsigned long n, unsigned long size )
{
	return calloc( n, size );
}
#endif

#ifndef OPENOCL_BASE_PSA_CRUNTIME_FREE
#define OPENOCL_BASE_PSA_CRUNTIME_FREE
void CRuntime_free( void* ptr )
{
	return free( ptr );
}
#endif


#ifdef __cplusplus
#ifdef TESTING


bool testFindRelativePath()
{
	bool result = 1;

	const char* expected = "../securizant/activate-1.0/bin/activate";
	const char* dir1     = "/opt/bin/activate";
	const char* dir2     = "/opt/securizant/activate-1.0/bin/activate";

	char* relative = find_relative_path( dir1, dir2 );
	{
		if ( 0 != strcmp( expected, relative ) )
		{
			result = 0;
		}
	}
	CRuntime_free( relative );

	return result;
}

bool testFindRelativePath2()
{
	bool result = 1;

	const char* expected = "../../";
	const char* dir1     = "/local/software/Commands/bison-1.35/";
	const char* dir2     = "/local/software/";

	char* relative = find_relative_path( dir1, dir2 );
	{
		if ( 0 != strcmp( expected, relative ) )
		{
			result = 0;
		}
	}
	CRuntime_free( relative );

	return result;
}

bool testFindTargetBase()
{
	bool result = 1;

	const char* target = "/opt/securizant/activate-1.0";

	char* base = find_target_base( target );
	CRuntime_free( base );

	return result;
}

bool testset()
{
	bool result = 1;
	
	result = result && Testing::runTest( "testFindRelativePath", testFindRelativePath );
	result = result && Testing::runTest( "testFindRelativePath2", testFindRelativePath2 );
	result = result && Testing::runTest( "testFindTargetBase", testFindTargetBase );
	result = result && Testing::runTest( "activate", activate );

	return result;
}

#endif // TESTING
#endif // __cplusplus

/*************************************************************************
 *   Definitions
 *************************************************************************/

int
main( int argc, char** argv )
{
	ARGC = argc;
	ARGV = argv;

#ifdef __cplusplus
	#ifdef TESTING

	return !Testing::runTestSet( "Activate Testset", testset );

	#endif
#endif

	return !activate();

}

bool activate()
{
	int index = process_arguments( ARGC, ARGV );
	if ( verbose ) print_arguments();

	char** packages = &ARGV[index];
	return activate_packages( ARGC - index, packages );
}

bool
activate_packages( int count, char** packages )
{
	int i;
	char* absolutePackagePath;
	for ( i=0; i < count; i++ )
	{
		if ( (absolutePackagePath = find_absolute_path( packages[i] )) )
		{
			if ( deactivate )
			{
				deactivate_package( absolutePackagePath );
			} else {
				activate_package( absolutePackagePath );
			}
			CRuntime_free( absolutePackagePath );
		}
	}

	return true;
}

bool
activate_package( const char* app )
{
	char* atp = find_target_base( app );
	printf( "Activating: %s\n", app );

	if ( all )
	{
		activate_directory( app, atp, "bin" );
		activate_directory( app, atp, "etc" );
		activate_directory( app, atp, "include" );
		activate_directory( app, atp, "sbin" );
		activate_directory( app, atp, "info" );
		activate_lib_directory( app, atp, "lib" );
		activate_man_directory( app, atp, "man" );
	}
	else if ( lib )
	{
		activate_directory( app, atp, "bin" );
		activate_directory( app, atp, "info" );
		activate_directory( app, atp, "sbin" );
		activate_lib_directory( app, atp, "lib" );
		activate_man_directory( app, atp, "man" );

		if ( etc ) activate_directory( app, atp, "etc" );
		if ( include ) activate_directory( app, atp, "include" );
	}
	else
	{
		activate_directory( app, atp, "bin" );
		activate_directory( app, atp, "sbin" );
		activate_directory( app, atp, "info" );
		activate_man_directory( app, atp, "man" );

		if ( etc ) activate_directory( app, atp, "etc" );
		if ( include ) activate_directory( app, atp, "include" );
	}

	CRuntime_free( atp );

	return true;
}

bool
activate_directory( const char* absolutePackagePath,
                    const char* absoluteTargetPath,
                    const char* directoryToActivate )
{
	char* src = concat_dirs( absolutePackagePath, directoryToActivate );

	DIR* directory;
	if ( NULL != (directory = opendir( src )) )
	{
		struct dirent* entry;
		while ( NULL != (entry = readdir( directory )) )
		{
			if ( entry->d_name[0] != '.' )
			{
				activate_file( absolutePackagePath, absoluteTargetPath, directoryToActivate, entry->d_name );
			}
		}
		closedir( directory );
	}
	CRuntime_free( src );

	return true;
}

bool
activate_lib_directory( const char* absolutePackagePath,
                        const char* absoluteTargetPath,
                        const char* directoryToActivate )
{
	char* src = concat_dirs( absolutePackagePath, directoryToActivate );
	char* tgt = concat_dirs( absoluteTargetPath, directoryToActivate );

	bool activate_pkgconfig = 0;

	DIR* directory;
	if ( NULL != (directory = opendir( src )) )
	{
		struct dirent* entry;
		while ( NULL != (entry = readdir( directory )) )
		{
			if ( entry->d_name[0] != '.' )
			{
				if ( 0 == strcmp( entry->d_name, "pkgconfig" ) )
				{
					activate_pkgconfig = 1;
				} else {
					activate_file( absolutePackagePath, absoluteTargetPath, directoryToActivate, entry->d_name );
				}
			}
		}
		closedir( directory );
	}

	if ( activate_pkgconfig )
	{
		activate_directory( src, tgt, "pkgconfig" );
	}

	CRuntime_free( src );
	CRuntime_free( tgt );

	return true;
}

bool
activate_man_directory( const char* absolutePackagePath,
                        const char* absoluteTargetPath,
                        const char* directoryToActivate )
{
	const char* man_dirs[9];
	man_dirs[0] = "man1";
	man_dirs[1] = "man2";
	man_dirs[2] = "man3";
	man_dirs[3] = "man4";
	man_dirs[4] = "man5";
	man_dirs[5] = "man6";
	man_dirs[6] = "man7";
	man_dirs[7] = "man8";
	man_dirs[8] = "man9";

	char* src = concat_dirs( absolutePackagePath, directoryToActivate );
	char* tgt = concat_dirs( absoluteTargetPath, directoryToActivate );

	int i=0;
	for ( i=0; i < 9; i++ )
	{
		char* man_src = concat_dirs( src, man_dirs[i] );
		{
			DIR* directory;
			if ( NULL != (directory = opendir( man_src )) )
			{
				struct dirent* entry;
				while ( NULL != (entry = readdir( directory )) )
				{
					if ( entry->d_name[0] != '.' )
					{
						activate_file( src, tgt, man_dirs[i], entry->d_name );
					}
				}
				closedir( directory );
			}
		}
		CRuntime_free( man_src );
	}

	CRuntime_free( src );
	CRuntime_free( tgt );

	return true;
}

bool
activate_file( const char* absolutePackagePath,
               const char* absoluteTargetPath,
               const char* directoryToActivate,
               const char* file )
{
	bool ret = true;

	char* src_dir = concat_dirs( absolutePackagePath, directoryToActivate );
	char* src_file = concat_dirs( src_dir, file );
	char* dest_dir = concat_dirs( absoluteTargetPath, directoryToActivate );
	char* dest_file = concat_dirs( dest_dir, file );

	char* relative = find_relative_path( dest_file, src_file );

	DIR* directory;
	if ( NULL == (directory = opendir( dest_dir )) )
	{
		mkdir( dest_dir, 0777 );
	} else {
		closedir( directory );
	}

	printf( "ln -s %s %s\n", relative, dest_file );
	if ( -1 == symlink( relative, dest_file ) )
	{
		printf( "error: %s\n", strerror( errno ) );
		ret = false;
	}

	CRuntime_free( src_dir );
	CRuntime_free( src_file );
	CRuntime_free( dest_dir );
	CRuntime_free( dest_file );
	CRuntime_free( relative );

	return ret;
}

bool
deactivate_package( const char* app )
{
	char* atp = find_target_base( app );
	printf( "Dectivating: %s\n", app );

	deactivate_directory( app, atp, "bin" );
	deactivate_directory( app, atp, "etc" );
	deactivate_directory( app, atp, "include" );
	deactivate_directory( app, atp, "sbin" );
	deactivate_directory( app, atp, "texinfo" );
	deactivate_lib_directory( app, atp, "lib" );
	deactivate_man_directory( app, atp, "man" );

	CRuntime_free( atp );

	return true;
}


bool
deactivate_directory( const char* absolutePackagePath,
                      const char* absoluteTargetPath,
                      const char* directoryToDeactivate )
{
	char* src = concat_dirs( absolutePackagePath, directoryToDeactivate );

	DIR* directory;
	if ( NULL != (directory = opendir( src )) )
	{
		struct dirent* entry;
		while ( NULL != (entry = readdir( directory )) )
		{
			if ( entry->d_name[0] != '.' )
			{
				deactivate_file( absolutePackagePath, absoluteTargetPath, directoryToDeactivate, entry->d_name );
			}
		}
		closedir( directory );
	}
	
	CRuntime_free( src );

	return true;
}

bool
deactivate_lib_directory( const char* absolutePackagePath,
                          const char* absoluteTargetPath,
                          const char* directoryToDeactivate )
{
	char* src = concat_dirs( absolutePackagePath, directoryToDeactivate );
	char* tgt = concat_dirs( absoluteTargetPath, directoryToDeactivate );

	bool deactivate_pkgconfig = 0;

	DIR* directory;
	if ( NULL != (directory = opendir( src )) )
	{
		struct dirent* entry;
			while ( NULL != (entry = readdir( directory )) )
			{
				if ( entry->d_name[0] != '.' )
				{
					if ( 0 == strcmp( entry->d_name, "pkgconfig" ) )
					{
						deactivate_pkgconfig = 1;
					} else {
						deactivate_file( absolutePackagePath, absoluteTargetPath, directoryToDeactivate, entry->d_name );
					}
				}
			}
		closedir( directory );
	}

	if ( deactivate_pkgconfig )
	{
		deactivate_directory( src, tgt, "pkgconfig" );
	}

	CRuntime_free( src );
	CRuntime_free( tgt );

	return true;
}

bool
deactivate_man_directory( const char* absolutePackagePath,
                          const char* absoluteTargetPath,
                          const char* directoryToActivate )
{
	const char* man_dirs[9];
	man_dirs[0] = "man1";
	man_dirs[1] = "man2";
	man_dirs[2] = "man3";
	man_dirs[3] = "man4";
	man_dirs[4] = "man5";
	man_dirs[5] = "man6";
	man_dirs[6] = "man7";
	man_dirs[7] = "man8";
	man_dirs[8] = "man9";

	char* src = concat_dirs( absolutePackagePath, directoryToActivate );
	char* tgt = concat_dirs( absoluteTargetPath, directoryToActivate );

	int i=0;
	for ( i=0; i < 9; i++ )
	{
		char* man_src = concat_dirs( src, man_dirs[i] );
		{
			DIR* directory;
			if ( NULL != (directory = opendir( man_src )) )
			{
				struct dirent* entry;
				while ( NULL != (entry = readdir( directory )) )
				{
					if ( entry->d_name[0] != '.' )
					{
						deactivate_file( src, tgt, man_dirs[i], entry->d_name );
					}
				}
				closedir( directory );
			}
		}
		CRuntime_free( man_src );
	}

	CRuntime_free( src );
	CRuntime_free( tgt );

	return true;
}

bool
deactivate_file( const char* absolutePackagePath,
                 const char* absoluteTargetPath,
                 const char* directoryToActivate,
                 const char* file )
{
	char* src_dir = concat_dirs( absolutePackagePath, directoryToActivate );
	char* src_file = concat_dirs( src_dir, file );
	char* dest_dir = concat_dirs( absoluteTargetPath, directoryToActivate );
	char* dest_file = concat_dirs( dest_dir, file );

	char* relative = find_relative_path( dest_file, src_file );

	int size = strlen( src_file ) + 1;
	char* buffer = (char*) CRuntime_calloc( sizeof( char ), size );

	if ( -1 != readlink( dest_file, buffer, size ) )
	{
		if ( 0 == strcmp( buffer, relative ) )
		{
			printf( "rm %s --> %s\n", dest_file, buffer );
			if ( -1 == unlink( dest_file ) )
			{
				printf( "activate: %s\n", strerror( errno ) );
			}
		}
	}

	CRuntime_free( buffer );

	CRuntime_free( relative );

	CRuntime_free( src_dir );
	CRuntime_free( src_file );
	CRuntime_free( dest_dir );
	CRuntime_free( dest_file );

	return true;
}

char*
find_absolute_path( const char* aRelativePath )
{
	char* absPath = null;
	char* cwd;

	cwd = find_cwd();

	if ( 0 == chdir( aRelativePath ) )
	{
		absPath = find_cwd();
		chdir( cwd );
	}
	CRuntime_free( cwd );
	return absPath;
}

char*
find_cwd()
{
	int capacity = 64;
	char* cwd = (char*) CRuntime_calloc( sizeof( char ), capacity ); 

	while ( NULL == getcwd( cwd, capacity ) )
	{
		CRuntime_free( cwd );
		capacity = capacity*2;
		cwd = (char*) CRuntime_malloc( sizeof( char ) * capacity );
	}
	return cwd;
}

char*
find_target_base( const char* anAbsolutePath )
{
	const char* previous = "..";

	char* target = null;

	char* cwd = find_cwd();
	{
		bool loop = true;

		chdir( anAbsolutePath );

		while ( (0 == chdir( previous )) && (true == loop) )
		{
			char* current = find_cwd();
			{
				printf( "Checking: %s\n", current );
				if ( 0 == chdir( "bin" ) )
				{
					printf( "Using: %s\n", current );
					chdir( previous );
					loop = false;
					target = find_cwd();
					loop = false;
				}

				if ( 0 == strcmp( current, "/" ) ) {
					loop = false;
				}
			}
			CRuntime_free( current );
		}
	}
	chdir( cwd );
	CRuntime_free( cwd );

	return target;
}

int
process_arguments( int argc, char** argv )
{
	int c;

	while ( -1 != (c = getopt( argc, argv, "adbeilmstv" )) )
	{
		switch ( c )
		{
		case 'a':
			all = true;
			break;
		case 'd':
			deactivate = true;
			break;
		case 'n':
			nothing = true;
			break;
		case 'b':
			bin = true;
			break;
		case 'e':
			etc = true;
			break;
		case 'i':
			include = true;
			break; 
		case 'l':
			lib = true;
			break;
		case 'm':
			man = true;
			break;
		case 's':
			sbin = true;
			break;
		case 't':
			texinfo = true;
			break;
		case 'v':
			verbose = true;
		}
	}

	return optind;
}

void
print_arguments()
{
	if ( all ) {
		printf( "--all\n" );
	}
	if ( deactivate ) {
		printf( "--deactivate\n" );
	}
	if ( nothing ) {
		printf( "--nothing\n" );
	}
	if ( bin ) {
		printf( "--bin\n" );
	}
	if ( etc ) {
		printf( "--etc\n" );
	}
	if ( include ) {
		printf( "--include\n" );
	}
	if ( lib ) {
		printf( "--lib\n" );
	}
	if ( man ) {
		printf( "--man\n" );
	}
	if ( sbin ) {
		printf( "--sbin\n" );
	}
	if ( texinfo ) {
		printf( "--texinfo\n" );
	}
}

char*
concat_dirs( const char* first, const char* second )
{
	int size1 = strlen( first );
	int size2 = strlen( second );
	int size = size1 + size2 + 2;

	char* result = (char*) CRuntime_calloc( sizeof( char ), size );
	strcat( result, first );
	strcat( result, "/" );
	strcat( result, second );

	return result;
}

char*
find_relative_path( const char* source, const char* target )
{
//  source = /local/optional/bin/activate
//  target = /local/optional/software/package_management/activate-x.y.z/bin/activate
//  relative = ../../software/package_management/activate-x.y.z/bin/activate

	char* relative = (char*) CRuntime_calloc( 256, sizeof( char ));
	{
		int stepback = 0;
		int i = 0;

		char* aSource = strdup( source );
		char* aTarget = strdup( target );
		{
			char* src = strtok_r( aSource, "/", &aSource );
			char* dst = strtok_r( aTarget, "/", &aTarget );

			while ( 0 == strcmp( src, dst ) )
			{
				src = strtok_r( NULL, "/", &aSource );
				dst = strtok_r( NULL, "/", &aTarget );
				
				if ( !src || !dst ) break;
			}

			while ( NULL != src )
			{
				stepback++;
				src = strtok_r( NULL, "/", &aSource );
			}

			{
				int end = strlen( source );
				if ( '/' != source[end - 1] )
				{
					stepback--;
				}
			}

			for ( i=0; i < stepback; i++ )
			{
				relative = strcat( relative, "../" );
			}

			while ( dst )
			{
				relative = strcat( relative, dst );
				dst = strtok_r( NULL, "/", &aTarget );

				if ( NULL != dst )
				{
					relative = strcat( relative, "/" );
				}
			}
		}
		// free( aSource ); // Causes segfault on linux
		// free( aTarget ); // Reimplement strtok_r
	}
	return relative;
}
